'use strict';

const request = require('request');
const _ = require('../util_modules');
var fs = require('fs');
var path = require('path');

class API {
	constructor({
		auth = true,
		resolveOnHTTPError = false,
		maxRetry = 3,
		env = 'Development',
		log = {}
	} = {}) {
		this.requestOpts = {
			method: 'GET',
			json: true,
			headers: {
			}
		};
		this.authNeeded = auth;
		this.resolveOnHTTPError = resolveOnHTTPError;
		this.logOpts = log;
		this.maxRetryCount = maxRetry;
		this.retryCount = 0;
		this.progress = null;
		this.downloadfile = false;
		this.zipPath = null;
	}

	_logReq() {
		let qs = this.requestOpts.qs
			? '\nQuery params: ' + JSON.stringify(this.requestOpts.qs)
			: '';
		qs = this.logOpts.skipQuery === true ? '\n<request query omitted>' : qs;

		let body = this.requestOpts.body
			? '\nRequest Body: ' + JSON.stringify(this.requestOpts.body)
			: '';
		body = this.logOpts.skipReqBody === true ? '\n<request body omitted>' : body;

		body += this.requestOpts.form
			? '\nRequest Form: ' + JSON.stringify(this.requestOpts.form)
			: '';
		_.LOGGER.debug(
			'>>>> HTTP REQUEST : ' +
				this.requestOpts.method +
				' ' +
				this.requestOpts.url +
				' ' +
				qs +
				body +
				'\n'
		);
	}

	_logResp(resp) {
		_.LOGGER.debug(
			'<<<< HTTP RESPONSE : ' +
				resp.statusCode +
				'\n' +
				'Response Headers' +
				JSON.stringify(resp.headers) +
				'\n'
		);
		if (!_.JS.isUndefined(resp.timings)) {
			_.LOGGER.debug('TOTAL TIME TAKEN : ' + Number(resp.timings.end.toFixed(3)) + ' ms\n');
			_.LOGGER.debug('REQUEST PHASES : ' + resp.timingPhases);
		}
		if (!_.JS.isUndefined(this.logOpts.progress)) {
			_.JS.set(this.logOpts, 'progress.total', resp.headers['content-length']);
			this.progress = require('../progress')(this.logOpts.progress);
		}
	}

	_logProgress(chunk) {
		if (!_.JS.isNull(this.progress)) {
			this.progress.tick(chunk.length);
		}
	}

	_addAuthHeader() {
		return require('./credential')
			.getAccessToken()
			.then((access_token) => {
				_.JS.set(this.requestOpts, 'headers.authorization', 'Bearer ' + access_token);
			});
	}

	_parseResponse(resp, body) {
		if (resp.statusCode >= 400 && !this.logOpts.skipRespBody) {
			_.LOGGER.debug('Response Body : ' + JSON.stringify(resp.body) + '\n');
			if (!this.resolveOnHTTPError) {
				return Promise.reject(require('../errorResponse')(resp, body));
			}
		}

		return Promise.resolve({
			status: resp.statusCode,
			response: resp,
			body:
				_.JS.includes(resp.headers['content-type'], 'json') && typeof body === 'string'
					? JSON.parse(body)
					: body
		});
	}

	_retry(err) {
		if (++this.retryCount > this.maxRetryCount) {
			return _.PROMISE.reject('API Error after retrying : ' + err.message, {
				original: err,
				exit: 2
			});
		}
		_.LOGGER.debug('API Error while firing request : ' + err.message + '\n');
		_.LOGGER.debug('>>>> RETRYING\n');
		return this._request(true);
	}

	_request(retry = false) {
		if (!retry) {
			this._logReq();
		}
		return new Promise((resolve, reject) => {
			request(this.requestOpts, (err) => {
				if (err) {
					switch (err.code) {
						case 'ETIMEDOUT':
						case 'ENOTFOUND':
						case 'ESOCKETTIMEDOUT':
							_.LOGGER.warning(
								err.code +
									' has occured while communicating remote server. Kindly be patient while the cli retries this request.'
							);
							return this._retry(err)
								.then((result) => {
									resolve(result);
								})
								.catch((err) => {
									_.LOGGER.error(err);
									process.exit(1);
								});
						default: 
								_.LOGGER.error('API Error. ' + err.message);
								process.exit(1);
					}
				}
			})
				.on('response', (resp) => {
					this._logResp(resp);
					if(this.downloadfile){
						 resp.pipe(fs.createWriteStream(path.join(this.zipPath,'app_file.zip')));
						 _.LOGGER.debug('Zip file is downloaded');
					}
				})
				.on('complete', (resp, body) => {
					return this._parseResponse(resp, body)
						.then((result) => {
							resolve(result);
						})
						.catch((err) => {
							_.LOGGER.error(err);
							process.exit(1);
						});
				})
				.on('data', (data) => {
					this._logProgress(data);
				});
		});
	}

	fire(method, path, options) {
		let dctype = _.JS.get(options,'dctype');
		let origin = this.urlDcEncoder(_.JS.get(options, 'origin'),dctype);
		let json =  _.JS.get(options, 'json', this.requestOpts.json);
		this.resolveOnHTTPError = _.JS.get(options, 'resolveOnError', this.resolveOnHTTPError);
		this.maxRetryCount = _.JS.get(options, 'maxRetry', this.maxRetryCount);
		this.authNeeded = _.JS.get(options, 'auth', this.authNeeded);
		this.logOpts = _.JS.get(options, 'log', this.logOpts);
		this.downloadfile = _.JS.get(options, 'downloadfile', this.downloadfile);
		this.zipPath = _.JS.get(options, 'zipPath', this.zipPath);
		this.requestOpts = _.JS.defaults(
			_.JS.omit(options, ['origin', 'auth', 'maxRetry', 'resolveOnError', 'log', 'dctype', 'json']),
			this.requestOpts
		);
		this.requestOpts.json = json;
		this.requestOpts.url = origin + path;
		this.requestOpts.method = method;
		if (this.authNeeded) {
			return this._addAuthHeader().then(() => {
				return this._request();
			});
		}
		return this._request();
	}

	get(path, options) {
		return this.fire('GET', path, options);
	}

	put(path, options) {
		return this.fire('PUT', path, options);
	}

	post(path, options) {
		return this.fire('POST', path, options);
	}

	delete(path, options) {
		return this.fire('DELETE', path, options);
	}
	urlDcEncoder(url,dctype){
		if(dctype){
			if(_.CONSTANTS.DCTYPES[dctype]){
				return url.replace('dctype',_.CONSTANTS.DCTYPES[dctype]);
			}else{
				_.LOGGER.debug("Invalid DC type so using US dc");
				return url.replace('dctype',_.CONSTANTS.DCTYPES["us"]);
			}
		}else{
			_.LOGGER.error('DC Information is not found.Kindly re-login using zet login --force');
		}
	}
}

module.exports = API;
