/*
Copyright (c) 2017, ZOHO CORPORATION
License: MIT
*/

//var fs = require('fs');
var path = require('path');
var i18next = require('i18next');
var CONST = require('../commands/constants.js');
var util = require('../utils');
const service = require('../utils/service.js');
var isString = util.isString;
var isUndefined = util.isUndefined;
var isObject = util.isObject;
var isArray = util.isArray;

const VALID_KEYS = [
    "name", "type", "placeholder", "defaultValue", "help"
];

const TYPE_VALUES = [
    "string", "integer", "float", "boolean", "Application", "Form", "Report", "Field"
];

function validatePluginManifest(projectRootDir) {
    var errors = [];
    var pluginManifest = require(path.join(projectRootDir, CONST.PLUGIN_MANIFEST));
    let manifestKeys = Object.keys(pluginManifest);

    //cspDomains exists validation
    if (pluginManifest.cspDomains) {
        let cspDomains = pluginManifest.cspDomains;
        errors = errors.concat(service.validateCspDomains(cspDomains));
    }

    //config exists validation
    var configs = pluginManifest.config;
    if (!isUndefined(configs)) {
        if (isArray(configs)) {
            if (configs.length > 20) {
                errors.push(i18next.t('desk_pluginmanifest_configlimitexceeded'));
            } else {
                errors = errors.concat(validateConfigOptions(configs));
            }
        } else {
            errors.push(i18next.t('creator_pluginmanifest_invalidconfig'));
        }
    }

    return errors;
}

function validateConfigOptions(configs) {
    var errors = [], nameArray = [];
    for (var i = 0; i < configs.length; i++) {
        var config = configs[i];
        if (!isObject(config)) {
            errors.push(i18next.t('creator_pluginmanifest_invalidconfigoption'));
            continue;
        }
        if( isObject(config) && Object.keys(config).length === 0 ) {
            errors.push(i18next.t('desk_pluginmanifest_emptyconfig'));
            continue;
        }

        for(var key in config){
            if(VALID_KEYS.indexOf(key) === -1) {
                errors.push(i18next.t('creator_pluginmanifest_invalidconfigkey'));
            }
        }

        var name = config.name.toLowerCase();
        var type = config.type;
        var placeholder = config.placeholder;
        var defaultValue = config.defaultValue;
        var help = config.help;

        if (isUndefined(name)) {
            errors.push(i18next.t('desk_pluginmanifest_missingconfigname'));
        } else {
            if (!isString(name)) {
                errors.push(i18next.t('desk_pluginmanifest_invalidconfigname'));
            } else if (name.trim() === "") {
                errors.push(i18next.t('creator_pluginmanifest_emptyconfigname'));
            } else if (!/^[a-zA-Z][a-zA-Z0-9-_ ]*$/i.test(name)) {
                errors.push(i18next.t("desk_pluginmanifest_invalidcharacters"));
            } else if(isString(name) && name.length > 50) {
                errors.push(i18next.t('creator_pluginmanifest_invalidnamelength'));
            } else if (isString(name) && nameArray.indexOf(name.trim()) >= 0) {
                errors.push(i18next.t("desk_pluginmanifest_duplicatename") + "'config' : " + config.name);
            } else {
                nameArray.push(name.trim());
            }
        }

        if (isUndefined(type)) {
            errors.push(i18next.t('creator_pluginmanifest_missingconfigtype'));
        } else {
            if (!isString(type)) {
                errors.push(i18next.t('projects_pluginmanifest_invalidtypeinconfig'));
            } else if (TYPE_VALUES.indexOf(type) === -1){
                errors.push(i18next.t('creator_pluginmanifest_invalidconfigtype'));
            }
        }

        if (!isUndefined(placeholder)) {
            if(!isString(placeholder)){
                errors.push(i18next.t('creator_pluginmanifest_invalidconfigplaceholder'));
            } else if (placeholder.length>40) {
                errors.push(i18next.t('creator_pluginmanifest_invalidplaceholderlength'));
            }
        }
        
        if (!isUndefined(defaultValue)) {
            if(!isString(defaultValue)){
                errors.push(i18next.t('creator_pluginmanifest_invaliddefaultvalue'));
            }
        }

        if (!isUndefined(help)) {
            if(!isString(help)){
                errors.push(i18next.t('creator_pluginmanifest_invalidhelp'));
            } else if (help.length>140) {
                errors.push(i18next.t('creator_pluginmanifest_invalidhelplength'));
            }
        }

    }
    return errors;
}

module.exports = {

    name: i18next.t('pluginmanifestrule_name'),
    run: function (projectRootDir, fileTree) {
        return validatePluginManifest(projectRootDir);
    }
}