/*
Copyright (c) 2017, ZOHO CORPORATION
License: MIT
*/

var fs = require('fs');
var path = require('path');
var i18next = require('i18next');
var CONST = require('../commands/constants.js');
var util = require('../utils');
var isString = util.isString;
var isUndefined = util.isUndefined;
var isObject = util.isObject;
var isArray = util.isArray;
var isBoolean = util.isBoolean;

var MANIFEST_KEY = [ "service", "type", "widgets" ];
var LOCATION_CONSTANT = [
    "dashboard","webtab","configuration","request.detail.menu","request.detail.subtab",
    "request.detail.rightpanel","request.form.rightpanel","change.detail.menu","change.detail.subtab",
    "change.stage.detail.subtab","change.detail.rightpanel","release.detail.menu","release.detail.subtab",
    "release.stage.detail.subtab","release.detail.rightpanel","asset.detail.menu","asset.detail.subtab",
    "asset.detail.rightpanel","problem.detail.menu","problem.detail.subtab","problem.detail.rightpanel","change.form.rightpanel"
];

function validatePluginManifest(projectRootDir) {

    var errors = [];
    var pluginManifest = require(path.join(projectRootDir, CONST.PLUGIN_MANIFEST));

    let manifestKeys = Object.keys(pluginManifest);

    //plugin manifest mandatory key validation
    for (var i = 0; i < MANIFEST_KEY.length; i++) {
        if (manifestKeys.indexOf(MANIFEST_KEY[i]) === -1) {
            errors.push("\"" + MANIFEST_KEY[i] + "\"" + i18next.t('pluginmanifest_keymissing'));
        }
    }

    //storage type check
    var type = pluginManifest.type;
    if (type !== undefined) {
        if (!isString(type) || ( isString(type) && type.trim() === "" ) ||
            (type !== 'personal')) {
            errors.push(i18next.t('desk_pluginmanifest_invalidtype'));
        }
    }

    //module validation
    var widgets = pluginManifest.widgets;
    if (!isUndefined(widgets)) {
        if (!isArray(widgets)) {
            errors.push(i18next.t('desk_pluginmanifest_invalidwidgettype'));
        } else if (widgets.length === 0) {
            errors.push(i18next.t('desk_pluginmanifest_widgetisempty'));
        } else if (!widgets.every((widget) => {return isObject(widget);})) {
            errors.push(i18next.t('sdp_pluginmanifest_invalidwidget'));
        } else {
            errors = errors.concat(validateWidgetConfigs(projectRootDir, widgets));
        }
    }

    return errors;

}

function validateWidgetConfigs(projectRootDir, widgets) {
    var errs = [], nameArray = [];

    for (var i = 0; i < widgets.length; i++) {
        var widget = widgets[i];
        var name = widget.name;
        if (isUndefined(name)) {
            errs.push(i18next.t('desk_pluginmanifest_missingwidgetname') + name);
        } else {
            if ( !isString(name) || ( isString(name) && name.trim() === "" ) ) {
                errs.push(i18next.t('desk_pluginmanifest_invalidwidgetname') + name);
            }
            if (!/^[a-zA-Z][a-zA-Z0-9-_ ]*$/i.test(name)) {
                errs.push(i18next.t("desk_pluginmanifest_invalidcharacters") + "'widgets' : " + name);
            } else if (isString(name) && nameArray.indexOf(name.trim()) >= 0) {
                errs.push(i18next.t("desk_pluginmanifest_duplicatename") + "'widgets' : " + name);
            } else if( isString(name) ) {
                nameArray.push(name.trim());
            
                var locations = widget.locations;
                var description = widget.description;
                var showWidgetHeadeer = widget.show_widget_header_in_dashboard;
                var pathURL = widget.path_url;
                
                if (isUndefined(locations)) {
                    errs.push(i18next.t('sdp_pluginmanifest_missinglocationinwidget') + name);
                } else if(!isArray(locations)){
                    errs.push(i18next.t('sdp_pluginmanifest_invalidlocation') + name);
                } else {
                    for(var j=0;j<locations.length;j++) {
                        if(checkValidLocations(locations[j])){
                            errs.push(locations[j]+ " - " + i18next.t('desk_pluginmanifest_invalidlocation') + name);
                        }
                    }
                }

                if (isUndefined(description)) {
                    errs.push(i18next.t('sdp_pluginmanifest_missingdescription') + name);
                } else if (!isString(description)) {
                    errs.push(i18next.t('sdp_pluginmanifest_invaliddescriptiontype') + name);
                }

                var pat = new RegExp('^(?:[a-z]+:)?//', 'i');
                if (isUndefined(pathURL)) {
                    errs.push(i18next.t('sdp_pluginmanifest_missingpathurl') + name);
                }  else if (!util.isValidURL(pathURL)) {
                    if (isString(pathURL)) {
                        var startsWithSlash = pathURL.indexOf('/') === 0;
                        if (startsWithSlash) {
                            if (!fs.existsSync(path.join(projectRootDir, pathURL))) {
                                errs.push(i18next.t('sdp_pluginmanifest_missingfileinappsource'));
                            }
                        } else {
                            errs.push(i18next.t('sdp_pluginmanifest_invalidrelativepathurl') + name);
                        }
                    } else {
                        errs.push(i18next.t('sdp_pluginmanifest_invalidpathurltype') + name);
                    }
                }

                if (!isUndefined(showWidgetHeadeer)) {
                    if (!isBoolean(showWidgetHeadeer)) {
                        errs.push(i18next.t('sdp_pluginmanifest_invalidwidgetheader') + name);
                    }
                }
            }
       
        }

    }
    return errs;
}

function checkValidLocations(location) {
    return LOCATION_CONSTANT.indexOf(location) === -1;
}

module.exports = {
    name: i18next.t('pluginmanifestrule_name'),
    run: function (projectRootDir, fileTree) {
        return validatePluginManifest(projectRootDir);
    }
}